package main

import (
	"fmt"
	"io/fs"
	"os"
	"os/exec"
	"os/user"
	"path/filepath"
	"strings"
	"sync"
	"time"

	vdirsyncer "git.sr.ht/~whynothugo/autovdirsyncer/vdirsyncer"
	"github.com/fsnotify/fsnotify"
)

const (
	passiveInterval = time.Minute * 15 // Period between automatic syncs if nothing happens.
	activeDebounce  = time.Second * 2  // Period to debounce when changes are happening.
)

var (
	vdirsyncerIsRunning sync.Mutex
)

func runSynchronization() {
	vdirsyncerIsRunning.Lock()

	cmd := exec.Command("vdirsyncer", "sync")
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr

	err := cmd.Run()
	if err != nil {
		fmt.Fprintf(os.Stderr, "Error running vdirsyncer: %v.\n", err.Error())

		if exitError, ok := err.(*exec.ExitError); ok {
			fmt.Fprintf(
				os.Stderr,
				"Got exit code %d from vdirsyncer: %v.\n",
				exitError.ExitCode(),
				err.Error(),
			)
		}
	} else {
		fmt.Println("vdirsyncer ran successfully.")
	}

	vdirsyncerIsRunning.Unlock()
}

func scheduleSynchronizations(timer *time.Timer) {
	for {
		<-timer.C
		runSynchronization()
		timer.Reset(passiveInterval)
	}
}

func listenForEvents(watcher *fsnotify.Watcher, timer *time.Timer) {
	for {
		select {
		case event, ok := <-watcher.Events:
			if !ok {
				return
			}
			// fsnotify only sends write-events, so we don't need
			// to check the event type (e.g. there are no "open" or
			// "read" events).

			fmt.Println("File altered:", event.Name)
			// XXX: If changes happen continuously within two
			// second intervals, sync never happens.
			timer.Reset(activeDebounce)
		case err, ok := <-watcher.Errors:
			if !ok {
				return
			}
			fmt.Println("error:", err)
		}
	}
}

func addDirsToWatcher(watcher *fsnotify.Watcher, dirsToWatch []string) error {
	home, err := user.Current()
	if err != nil {
		return err
	}

	for _, dir := range dirsToWatch {
		// Expand `~`.
		// TODO: Do this in vdirsyncer.
		if strings.HasPrefix(dir, "~/") {
			dir = filepath.Join(home.HomeDir, dir[2:])
		}

		err := filepath.WalkDir(dir, func(path string, d fs.DirEntry, err error) error {
			if err != nil {
				fmt.Fprintf(os.Stderr, "Failed to read path %q: %v.\n", path, err)
				return err
			}
			if d.IsDir() {
				fmt.Printf("Listening for changes on %+v \n", path)
				err = watcher.Add(path)
				if err != nil {
					fmt.Println(err)
					os.Exit(1)
				}
			}
			return nil
		})
		if err != nil {
			return err
		}
	}

	return nil
}

func main() {
	// Watcher that monitors for local fs changes.
	watcher, err := fsnotify.NewWatcher()
	if err != nil {
		fmt.Println(err)
		os.Exit(3)
	}
	defer watcher.Close()

	// Get the paths we'll monitor.
	storagePaths, err := vdirsyncer.GetVdirsyncerStoragePaths()
	if err != nil {
		fmt.Println(err)
		os.Exit(2)
	}

	err = addDirsToWatcher(watcher, storagePaths)
	if err != nil {
		fmt.Println(err)
		os.Exit(3)
	}

	timer := time.NewTimer(passiveInterval)

	go listenForEvents(watcher, timer)
	go scheduleSynchronizations(timer)

	// Never exit.
	select {}
}
